/*
   Copyright 2005 Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
    * Neither the name of Red Hat, Inc., nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "../config.h"
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dbus/dbus.h>
#include "common.h"
#include "mainloop.h"

#ident "$Id: mainloop.c,v 1.3 2005/06/27 23:30:21 nalin Exp $"

static struct watch_list {
	DBusWatch *watch;
	struct watch_list *next;
} *watch_list;

static struct timeout_list {
	DBusTimeout *timeout;
	struct timeval start;
	struct timeout_list *next;
} *timeout_list;

/* Add a watch item to the list.  If it's already there, don't add it again. */
static dbus_bool_t
watch_add(DBusWatch *watch, void *data)
{
	struct watch_list *l;
	for (l = watch_list; l != NULL; l = l->next) {
		if (l->watch == watch) {
			return TRUE;
		}
	}
	l = malloc(sizeof(struct watch_list));
	if (l == NULL) {
		return FALSE;
	}
	memset(l, 0, sizeof(*l));
	l->watch = watch;
	l->next = watch_list;
	watch_list = l;
	return TRUE;
}

/* Remove a watch item from the list. */
static void
watch_remove(DBusWatch *watch, void *data)
{
	struct watch_list *l, *pred;
	for (l = watch_list, pred = NULL; l != NULL; l = l->next) {
		if (l->watch == watch) {
			if (pred == NULL) {
				watch_list = l->next;
				free(l);
			} else {
				pred->next = l->next;
				free(l);
			}
			break;
		}
		pred = l;
	}
}

/* Toggle whether or not a watch item is enabled. */
static void
watch_toggle(DBusWatch *watch, void *data)
{
}

/* Add a timeout item to the timeout list. */
static dbus_bool_t
timeout_add(DBusTimeout *timeout, void *data)
{
	struct timeout_list *l;
	struct timeval tv;
	for (l = timeout_list; l != NULL; l = l->next) {
		if (l->timeout == timeout) {
			return TRUE;
		}
	}
	if (gettimeofday(&tv, NULL) == -1) {
		/* aargh! */
		return FALSE;
	}
	l = malloc(sizeof(struct timeout_list));
	if (l == NULL) {
		return FALSE;
	}
	memset(l, 0, sizeof(*l));
	l->timeout = timeout;
	memcpy(&l->start, &tv, sizeof(l->start));
	l->next = timeout_list;
	timeout_list = l;
	return TRUE;
}

/* Remove a timeout item from the timeout list. */
static void
timeout_remove(DBusTimeout *timeout, void *data)
{
	struct timeout_list *l, *pred;
	for (l = timeout_list, pred = NULL; l != NULL; l = l->next) {
		if (l->timeout == timeout) {
			if (pred == NULL) {
				timeout_list = l->next;
				free(l);
			} else {
				pred->next = l->next;
				free(l);
			}
			break;
		}
		pred = l;
	}
}

/* Toggle whether or not a timeout is enabled. */
static void
timeout_toggle(DBusTimeout *timeout, void *data)
{
}

/* Set up arguments for select() using data from the watch and timeout lists. */
static void
prepare(int *maxfd, fd_set *rfds, fd_set *wfds, fd_set *efds,
	dbus_bool_t *use_tv, struct timeval *tv)
{
	struct watch_list *wl;
	struct timeout_list *tl;
	int fd, flags, current_interval, this_interval;
	struct timeval now;

	FD_ZERO(rfds);
	FD_ZERO(wfds);
	FD_ZERO(efds);
	wl = watch_list;
	*maxfd = -1;
	while (wl != NULL) {
		if (dbus_watch_get_enabled(wl->watch)) {
			fd = dbus_watch_get_fd(wl->watch);
			flags = dbus_watch_get_flags(wl->watch);
			if (flags & (DBUS_WATCH_READABLE | DBUS_WATCH_HANGUP)) {
				FD_SET(fd, rfds);
				*maxfd = (*maxfd > fd ? *maxfd : fd);
			}
			if (flags & DBUS_WATCH_WRITABLE) {
				FD_SET(fd, wfds);
				*maxfd = (*maxfd > fd ? *maxfd : fd);
			}
			if (flags & DBUS_WATCH_ERROR) {
				FD_SET(fd, efds);
				*maxfd = (*maxfd > fd ? *maxfd : fd);
			}
		}
		wl = wl->next;
	}
	*use_tv = FALSE;
	memset(tv, 0, sizeof(*tv));
	tl = timeout_list;
	while (tl != NULL) {
		*use_tv = (gettimeofday(&now, NULL) == 0);
		if (*use_tv && dbus_timeout_get_enabled(tl->timeout)) {
			current_interval = tv->tv_sec * 1000 +
					   tv->tv_usec / 1000;
			this_interval = dbus_timeout_get_interval(tl->timeout);
			this_interval -= 1000 * (now.tv_sec - tl->start.tv_sec);

			if ((tl == timeout_list) ||
			    (this_interval < current_interval)) {
				tv->tv_sec = this_interval / 1000;
				tv->tv_usec = (this_interval % 1000) * 1000;
			}
		}
		tl = tl->next;
	}
}

/* For each descriptor with its bit set in one of the fd_set items, call
 * dbus_watch_handle with the right flag to process any pending I/O. */
static void
handle(fd_set *rfds, fd_set *wfds, fd_set *efds, struct timeval *tv)
{
	struct watch_list *wl;
	struct timeout_list *tl;
	int fd, flags;
	wl = watch_list;
	fd = -1;
	while (wl != NULL) {
		if (dbus_watch_get_enabled(wl->watch)) {
			fd = dbus_watch_get_fd(wl->watch);
			flags = 0;
			if (FD_ISSET(fd, rfds)) {
				flags |= DBUS_WATCH_READABLE;
			}
			if (FD_ISSET(fd, wfds)) {
				flags |= DBUS_WATCH_WRITABLE;
			}
			if (FD_ISSET(fd, efds)) {
				flags |= DBUS_WATCH_ERROR;
			}
			if (flags != 0) {
				dbus_watch_handle(wl->watch, flags);
			}
		}
		wl = wl->next;
	}
	/* PROBABLY WRONG */
	if (fd != -1) {
		tl = timeout_list;
		while (tl != NULL) {
			if (dbus_timeout_get_enabled(tl->timeout)) {
				dbus_timeout_handle(tl->timeout);
			}
			tl = tl->next;
		}
	}
}

/* Connect our callbacks for the D-BUS connection. */
void
mainloop_init(DBusConnection *conn)
{
	dbus_connection_set_watch_functions(conn,
					    watch_add,
					    watch_remove,
					    watch_toggle,
					    NULL,
					    NULL);
	dbus_connection_set_timeout_functions(conn,
					      timeout_add,
					      timeout_remove,
					      timeout_toggle,
					      NULL,
					      NULL);
}

void
mainloop_done(DBusConnection *conn)
{
	dbus_connection_set_watch_functions(conn,
					    NULL,
					    NULL,
					    NULL,
					    NULL,
					    NULL);
	dbus_connection_set_timeout_functions(conn,
					      NULL,
					      NULL,
					      NULL,
					      NULL,
					      NULL);
}

/* Process I/O for the connection, dispatch messages, and then flush any
 * pending output.  Return the number of descriptors which we serviced.  Note
 * that if we got a timeout, 0 is a *successful* return code. */
int
mainloop_iterate(DBusConnection *conn)
{
	int i, max_fd;
	dbus_bool_t use_tv;
	fd_set rfds, efds, wfds;
	struct timeval tv;

	prepare(&max_fd, &rfds, &wfds, &efds, &use_tv, &tv);
	i = select(max_fd + 1, &rfds, &wfds, &efds, use_tv ? &tv : NULL);
	if (i != -1) {
		handle(&rfds, &wfds, &efds, &tv);
		while (dbus_connection_get_dispatch_status(conn) ==
		       DBUS_DISPATCH_DATA_REMAINS) {
			dbus_connection_dispatch(conn);
		}
#if DBUS_CHECK_VERSION(0,30)
		while (dbus_connection_has_messages_to_send(conn)) {
			dbus_connection_flush(conn);
		}
#elif DBUS_CHECK_VERSION(0,20)
		while (dbus_connection_get_outgoing_size(conn) > 0) {
			dbus_connection_flush(conn);
		}
#else
#error "Don't know how to check if messages need to be flushed!"
#endif
	}
	return i;
}

/* Set the handler for an arbitrary set of signals to SIG_DFL. */
void
mainloop_reset_signal_handlers(void)
{
#ifdef SIGALRM
	signal(SIGALRM, SIG_DFL);
#endif
#ifdef SIGHUP
	signal(SIGHUP, SIG_DFL);
#endif
#ifdef SIGINT
	signal(SIGINT, SIG_DFL);
#endif
#ifdef SIGQUIT
	signal(SIGQUIT, SIG_DFL);
#endif
#ifdef SIGILL
	signal(SIGILL, SIG_DFL);
#endif
#ifdef SIGABRT
	signal(SIGABRT, SIG_DFL);
#endif
#ifdef SIGKILL
	signal(SIGKILL, SIG_DFL);
#endif
#ifdef SIGSEGV
	signal(SIGSEGV, SIG_DFL);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE, SIG_IGN);
#endif
#ifdef SIGTERM
	signal(SIGTERM, SIG_DFL);
#endif
#ifdef SIGCHLD
	signal(SIGCHLD, SIG_DFL);
#endif
#ifdef SIGCONT
	signal(SIGCONT, SIG_DFL);
#endif
#ifdef SIGSTOP
	signal(SIGSTOP, SIG_DFL);
#endif
}
